//! Error handling.

pub use failure::{Error, Fail};

use std::fmt;

macro_rules! ok {
   ($v:expr, $msg:expr) => {
      match $v {
         Some(v) => Ok(v),
         None    => Err($crate::err::err_msg($msg)),
      }
   };
}

macro_rules! flag_ok {
   ($t:ident$(::$tc:ident)*, $v:expr) => {
      {
         let v = $v;

         match $t$(::$tc)*::from_bits(v) {
            Some(v) => Ok(v),
            None    => Err($crate::err::ReprError::new(stringify!($t), v)),
         }
      }
   };
}

macro_rules! bail {
   ($e:expr) => {
      return Err($crate::err::err_msg($e));
   };
}

#[macro_export]
macro_rules! backtrace {
   ($e:expr) => {
      if cfg!(debug_assertions) {
         dbg!($e.backtrace());
      }
   }
}

/// Returns an `Error` with a static string.
///
/// # Examples
///
/// ```
/// use maraiah::err::err_msg;
///
/// assert_eq!(format!("{}", err_msg("oh no not things")),
///            "oh no not things");
/// ```
pub fn err_msg(msg: &'static str) -> Error {ErrMsg(msg).into()}

impl ReprError
{
   /// Returns an `Error` with a message for representation errata.
   ///
   /// # Examples
   ///
   /// ```
   /// use maraiah::err::repr_error;
   ///
   /// assert_eq!(format!("{}", repr_error("TypeName", 77)),
   ///            "bad TypeName (77)");
   /// ```
   pub fn new<T: Into<i64>>(t: &'static str, n: T) -> Self
   {
      Self(t, n.into())
   }
}

impl Fail for ReprError {}
impl Fail for ErrMsg {}

impl fmt::Display for ReprError
{
   fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
   {
      write!(f, "bad {} ({})", self.0, self.1)
   }
}

impl fmt::Display for ErrMsg
{
   fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
   {
      f.write_str(self.0)
   }
}

/// A representation error for an integer.
#[derive(Debug, Eq, Ord, PartialEq, PartialOrd)]
pub struct ReprError(&'static str, i64);

#[derive(Debug)]
struct ErrMsg(&'static str);

/// A generic `failure` based `Result` type.
pub type ResultS<T> = Result<T, Error>;

// EOF
